;*******************************************
;* Dimmer Control                          *
;* (c) 2002 Martin Pitt (piware@piware.de) *
;*******************************************

.include "4414def.inc"

.equ	BAUDRATE	= 25	; 9600 baud on 4 MHz (0.2% error)

;* Interrupt vector table *************
	rjmp	Init		; Reset
	reti			; IRQ0
	reti			; IRQ1
	reti			; Timer 1 Capture
	rjmp	Phase		; Timer 1 Compare Match A
	reti			; Timer 1 Compare Match B
	reti			; Timer 1 Overflow
	reti			; Timer 0 Overflow
	reti			; Serial transfer complete
	rjmp	Receive		; UART, Rx complete
	reti			; UART Data register empty
	reti			; UART, Tx complete
	rjmp	ZeroCrossing	; Analog comparator

	;; global register allocation

.def PHASE	= r25 		; counter (0..99) within the current
				; A/C semiwave
.def ENMASK	= r26		; enable mask
.def ENABLED	= r27		; currently enabled outputs
	;; ZH must always be zero!

;;; memory layout
;;; $60 - $67	phase start table for outputs 0 to 7
;;; $68 - $87	conversion table: linear value -> phase start
	
;;; ***************************************************
;;; This IRQ is triggered on a zero crossing of the A/C.
;;; It synchronizes the timer 1 events to the A/C.
;;; ***************************************************
ZeroCrossing:
	cbi	ACSR, ACIE	; disable AC Interrupt to
				; filter out double interrupts

	;; clear triac triggers
	ldi	ENABLED, 0
	out	PORTA, ENABLED

	;; clear and start timer 1 and clear phase
	ldi	PHASE, (1<<CTC1)|(1<<CS10)
	out	TCCR1B, PHASE	;  enable timer with CK/1+CTC1
	ldi	PHASE, 0
	out	TCNT1H, PHASE
	out	TCNT1L, PHASE

	reti

;;; ***************************************************
;;; This IRQ is triggered every 1/100 of an A/C semiwave.
;;; It fires the triac of every output that has reached
;;; its start phase.
;;; ***************************************************
Phase:
	push	r1
	push	r16
	push	ZL
	in	r1, SREG

	clr	ENABLED

	;; compare loop
	ldi	ZL, $60
	ldi	ENMASK, 1

Phase_NextOut:
	ld	r16, Z+
	cp	PHASE, r16
	brlo	Phase_NotReached
	or	ENABLED, ENMASK

Phase_NotReached:
	lsl	ENMASK
	brne	Phase_NextOut

	out	PORTA, ENABLED

	;; increment and wrap phase
	inc	PHASE

	cpi	PHASE, 20
	brne	Phase_Exit

	sbi	ACSR, ACIE	; reenable AC Interrupt

Phase_Exit:
	out	SREG, r1
	pop	ZL
	pop	r16
	pop	r1
	reti

;;; ***************************************************
;;; This IRQ is triggered on a received char from the UART
;;; ***************************************************
Receive:
	push	r1
	in	r1, SREG
	push	r16
	push	r17
	push	ZL
	
	in	r16, UDR
	
	;; mask value and convert to start phase
	mov	ZL, r16
	andi	ZL, $1F
	subi	ZL, -$68
	ld	r17, Z		;  do lookup in conv table ($68+value)

	;; write phase start to table ($60 + r16[7..5])
	clc
	lsr	r16
	lsr	r16
	lsr	r16
	lsr	r16
	lsr	r16
	mov	ZL, r16
	subi	ZL, -$60
	st	Z, r17

exit:
	pop	ZL
	pop	r17
	pop	r16
	out	SREG, r1
	pop	r1
	reti

;;; ***************************************************
;;; Reset vector (main program)
;;; ***************************************************
Init:
	; init stack pointer
	ldi	r16,low(RAMEND) 		
	out	spl,r16
	ldi	r16,high(RAMEND)
	out	sph,r16

	rcall	InitConversionTable

	ser	r16		; Port A as output
	out	DDRA, r16

	clr	ENABLED
	clr	PHASE
	clr	ZH

	;; init interrupts
	sbi	ACSR, ACIE	;  enable AC Interrupt

	ldi	r16, (1<<OCIE1A)
	out	TIMSK, r16	;  enable timer 1 Compare A IRQ

	ldi	r16, high(399)
	ldi	r17, low(399)
	out	OCR1AH, r16
	out	OCR1AL, r17

	sbi	UCR, RXEN	;  initialize UART
	sbi	UCR, RXCIE
	ldi	r16, BAUDRATE
	out	UBRR, r16
	
	sei			;  enable global interrupts

Init_Loop:
	rjmp	Init_Loop

;;; ***************************************************
;;; Write conversion lookup table (linear value -> phase start)
;;; linear values range from 0..31, phase starts from 0 to 99
;;; to memory $68 to $87
;;; ***************************************************
InitConversionTable:
	ldi	ZL, $60

	; clear phase start table
	ldi	r16, 99
	st	Z+, r16
	st	Z+, r16
	st	Z+, r16
	st	Z+, r16
	st	Z+, r16
	st	Z+, r16
	st	Z+, r16
	st	Z+, r16

	; write conversion table
	ldi	r16, 100
	st	Z+, r16
	ldi	r16, 89
	st	Z+, r16
	ldi	r16, 84
	st	Z+, r16
	ldi	r16, 80
	st	Z+, r16
	ldi	r16, 77
	st	Z+, r16
	ldi	r16, 74
	st	Z+, r16
	ldi	r16, 71
	st	Z+, r16
	ldi	r16, 68
	st	Z+, r16
	ldi	r16, 66
	st	Z+, r16
	ldi	r16, 64
	st	Z+, r16
	ldi	r16, 62
	st	Z+, r16
	ldi	r16, 59
	st	Z+, r16
	ldi	r16, 57
	st	Z+, r16
	ldi	r16, 55
	st	Z+, r16
	ldi	r16, 53
	st	Z+, r16
	ldi	r16, 51
	st	Z+, r16
	ldi	r16, 49
	st	Z+, r16
	ldi	r16, 47
	st	Z+, r16
	ldi	r16, 45
	st	Z+, r16
	ldi	r16, 43
	st	Z+, r16
	ldi	r16, 41
	st	Z+, r16
	ldi	r16, 38
	st	Z+, r16
	ldi	r16, 36
	st	Z+, r16
	ldi	r16, 34
	st	Z+, r16
	ldi	r16, 32
	st	Z+, r16
	ldi	r16, 29
	st	Z+, r16
	ldi	r16, 26
	st	Z+, r16
	ldi	r16, 23
	st	Z+, r16
	ldi	r16, 20
	st	Z+, r16
	ldi	r16, 16
	st	Z+, r16
	ldi	r16, 11
	st	Z+, r16
	ldi	r16, 0
	st	Z+, r16

	ret
